/*
Copyright 2022 The KCP Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package proxy

import (
	"context"
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"io/ioutil"
	"net/http"
	"net/http/httputil"
	"net/url"

	"k8s.io/apimachinery/pkg/util/runtime"
	userinfo "k8s.io/apiserver/pkg/authentication/user"
	"k8s.io/apiserver/pkg/endpoints/request"
)

func newTransport(clientCert, clientKeyFile, caFile string) (*http.Transport, error) {
	caCert, err := ioutil.ReadFile(caFile)
	if err != nil {
		return nil, fmt.Errorf("failed to read CA file %q: %w", caFile, err)
	}

	caCertPool := x509.NewCertPool()
	caCertPool.AppendCertsFromPEM(caCert)

	cert, err := tls.LoadX509KeyPair(clientCert, clientKeyFile)
	if err != nil {
		return nil, fmt.Errorf("failed to load client certificate %q or key %q: %w", clientCert, clientKeyFile, err)
	}

	transport := http.DefaultTransport.(*http.Transport).Clone()
	transport.TLSClientConfig = &tls.Config{
		Certificates: []tls.Certificate{cert},
		RootCAs:      caCertPool,
	}

	return transport, nil
}

// WithProxyAuthHeaders does client cert termination by extracting the user and groups and
// passing them through access headers to the shard.
func WithProxyAuthHeaders(delegate http.HandlerFunc, UserHeader, GroupHeader string) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		if u, ok := request.UserFrom(r.Context()); ok {
			appendClientCertAuthHeaders(r.Header, u, UserHeader, GroupHeader)
		}

		delegate.ServeHTTP(w, r)
	}
}

func appendClientCertAuthHeaders(header http.Header, user userinfo.Info, UserHeader, GroupHeader string) {
	header.Set(UserHeader, user.GetName())

	for _, group := range user.GetGroups() {
		header.Add(GroupHeader, group)
	}
}

func newShardReverseProxy() *httputil.ReverseProxy {
	director := func(req *http.Request) {
		shardURL := ShardURLFrom(req.Context())
		if shardURL == nil {
			// should not happen if wiring is correct
			runtime.HandleError(fmt.Errorf("no shard URL found in request context"))
			req.URL.Scheme = "https"
			req.URL.Host = "notfound"
			return
		}

		req.URL.Scheme = shardURL.Scheme
		req.URL.Host = shardURL.Host
	}
	return &httputil.ReverseProxy{Director: director}
}

type shardKey int

const shardContextKey shardKey = iota

func WithShardURL(parent context.Context, shardURL *url.URL) context.Context {
	return context.WithValue(parent, shardContextKey, shardURL)
}

func ShardURLFrom(ctx context.Context) *url.URL {
	shardURL, ok := ctx.Value(shardContextKey).(*url.URL)
	if !ok {
		return nil
	}
	return shardURL
}
